@@ -31,8 +31,8 @@ use crate::connection::Connection;
 use crate::cookie::VoidCookie;
 use crate::errors::{ConnectionError, ParseError, ReplyError};
 use crate::protocol::xproto::{
-    Drawable, Format, Gcontext, GetImageReply, GetImageRequest, ImageFormat,
-    ImageOrder as XprotoImageOrder, PutImageRequest, Setup, VisualClass, Visualtype,
+    get_image, put_image, Drawable, Format, Gcontext, GetImageReply, ImageFormat,
+    ImageOrder as XprotoImageOrder, Setup, VisualClass, Visualid, Visualtype,
 };
 
 /// The description of a single color component.
@@ -466,9 +466,9 @@ number_enum! {
 #[derive(Debug, Copy, Clone, Eq, PartialEq, Hash)]
 pub enum ImageOrder {
     /// Least significant byte first
-    LSBFirst,
+    LsbFirst,
     /// Most significant byte first
-    MSBFirst,
+    MsbFirst,
 }
 
 impl TryFrom<XprotoImageOrder> for ImageOrder {
@@ -476,15 +476,15 @@ impl TryFrom<XprotoImageOrder> for ImageOrder {
 
     fn try_from(value: XprotoImageOrder) -> Result<Self, ParseError> {
         match value {
-            XprotoImageOrder::LSB_FIRST => Ok(Self::LSBFirst),
-            XprotoImageOrder::MSB_FIRST => Ok(Self::MSBFirst),
+            XprotoImageOrder::LSB_FIRST => Ok(Self::LsbFirst),
+            XprotoImageOrder::MSB_FIRST => Ok(Self::MsbFirst),
             _ => Err(ParseError::InvalidValue),
         }
     }
 }
 
 /// The description of an image.
-#[derive(Debug, PartialEq, Eq)]
+#[derive(Debug, PartialEq, Eq, Clone)]
 pub struct Image<'a> {
     /// Width in pixels.
     width: u16,
@@ -510,23 +510,6 @@ pub struct Image<'a> {
     data: Cow<'a, [u8]>,
 }
 
-impl<'a> ToOwned for Image<'a> {
-    // TODO: Can this somehow be turned into Image<'static>?
-    type Owned = Self;
-
-    fn to_owned(&self) -> Self::Owned {
-        Image {
-            width: self.width,
-            height: self.height,
-            scanline_pad: self.scanline_pad,
-            depth: self.depth,
-            bits_per_pixel: self.bits_per_pixel,
-            byte_order: self.byte_order,
-            data: self.data.to_owned(),
-        }
-    }
-}
-
 impl<'a> Image<'a> {
     /// The width in pixels.
     pub fn width(&self) -> u16 {
@@ -679,7 +662,8 @@ impl<'a> Image<'a> {
     /// Get an image from the X11 server.
     ///
     /// This function sends a [`GetImage`](crate::protocol::xproto::GetImageRequest) request, waits
-    /// for its response and wraps it in a new `Image`.
+    /// for its response and wraps it in a new `Image`. The image and the corresponding visual id
+    /// are returned.
     ///
     /// The returned image contains the rectangle with top left corner `(x, y)` and size `(width,
     /// height)` of the given `drawable`.
@@ -690,19 +674,21 @@ impl<'a> Image<'a> {
         y: i16,
         width: u16,
         height: u16,
-    ) -> Result<Self, ReplyError> {
-        let reply = GetImageRequest {
+    ) -> Result<(Self, Visualid), ReplyError> {
+        let reply = get_image(
+            conn,
+            ImageFormat::Z_PIXMAP,
             drawable,
             x,
             y,
             width,
             height,
-            format: ImageFormat::Z_PIXMAP,
-            plane_mask: !0,
-        }
-        .send(conn)?
+            !0,
+        )?
         .reply()?;
-        Ok(Self::get_from_reply(&conn.setup(), width, height, reply)?)
+        let visual = reply.visual;
+        let image = Self::get_from_reply(conn.setup(), width, height, reply)?;
+        Ok((image, visual))
     }
 
     /// Construct an `Image` from a `GetImageReply`.
@@ -723,7 +709,7 @@ impl<'a> Image<'a> {
             format.scanline_pad.try_into()?,
             reply.depth,
             format.bits_per_pixel.try_into()?,
-            ImageOrder::MSBFirst,
+            setup.image_byte_order.try_into()?,
             Cow::Owned(reply.data),
         )
     }
@@ -736,6 +722,11 @@ impl<'a> Image<'a> {
     /// The server's maximum request size is honored. This means that a too large `PutImage`
     /// request is automatically split up into smaller pieces. Thus, if this function returns an
     /// error, the image could already be partially sent.
+    ///
+    /// Before uploading, the image is translated into the server's native format via
+    /// [`Image::native`]. This may convert the image to another format, which can be slow. If you
+    /// intend to upload the same image multiple times, it is likely more efficient to call
+    /// [`Image::native`] once initially so that the conversion is not repeated on each upload.
     pub fn put<'c, Conn: Connection>(
         &self,
         conn: &'c Conn,
@@ -743,6 +734,18 @@ impl<'a> Image<'a> {
         gc: Gcontext,
         dst_x: i16,
         dst_y: i16,
+    ) -> Result<Vec<VoidCookie<'c, Conn>>, ConnectionError> {
+        self.native(conn.setup())?
+            .put_impl(conn, drawable, gc, dst_x, dst_y)
+    }
+
+    fn put_impl<'c, Conn: Connection>(
+        &self,
+        conn: &'c Conn,
+        drawable: Drawable,
+        gc: Gcontext,
+        dst_x: i16,
+        dst_y: i16,
     ) -> Result<Vec<VoidCookie<'c, Conn>>, ConnectionError> {
         // Upload the image without exceeding the server's maximum request size
         let max_bytes = conn.maximum_request_bytes();
@@ -760,19 +763,19 @@ impl<'a> Image<'a> {
             let next_lines = lines_per_request.min(self.height - y_offset);
             let next_byte_offset = byte_offset + usize::from(next_lines) * stride;
             let data = &self.data[byte_offset..next_byte_offset];
-            let request = PutImageRequest {
-                format: ImageFormat::Z_PIXMAP,
+            result.push(put_image(
+                conn,
+                ImageFormat::Z_PIXMAP,
                 drawable,
                 gc,
-                width: self.width,
-                height: next_lines,
+                self.width,
+                next_lines,
                 dst_x,
-                dst_y: dst_y + i16::try_from(y_offset).unwrap(),
-                left_pad: 0, // Must always be 0 for ZPixmap
-                depth: self.depth,
-                data: Cow::Borrowed(&data),
-            };
-            result.push(request.send(conn)?);
+                dst_y + i16::try_from(y_offset).unwrap(),
+                0, // left_pad must always be 0 for ZPixmap
+                self.depth,
+                data,
+            )?);
 
             y_offset += next_lines;
             byte_offset = next_byte_offset;
@@ -879,7 +882,7 @@ impl<'a> Image<'a> {
             BitsPerPixel::B4 => {
                 let mut pixel = pixel & 0x0f;
                 let odd_x = x % 2 == 1;
-                let mask = if odd_x == (self.byte_order == ImageOrder::MSBFirst) {
+                let mask = if odd_x == (self.byte_order == ImageOrder::MsbFirst) {
                     pixel <<= 4;
                     0xf0
                 } else {
@@ -890,16 +893,16 @@ impl<'a> Image<'a> {
             BitsPerPixel::B8 => data[row_start + x] = pixel as u8,
             BitsPerPixel::B16 => {
                 let (p0, p1) = match self.byte_order {
-                    ImageOrder::LSBFirst => (pixel, pixel >> 8),
-                    ImageOrder::MSBFirst => (pixel >> 8, pixel),
+                    ImageOrder::LsbFirst => (pixel, pixel >> 8),
+                    ImageOrder::MsbFirst => (pixel >> 8, pixel),
                 };
                 data[row_start + 2 * x + 1] = p1 as u8;
                 data[row_start + 2 * x] = p0 as u8;
             }
             BitsPerPixel::B24 => {
                 let (p0, p1, p2) = match self.byte_order {
-                    ImageOrder::LSBFirst => (pixel, pixel >> 8, pixel >> 16),
-                    ImageOrder::MSBFirst => (pixel >> 16, pixel >> 8, pixel),
+                    ImageOrder::LsbFirst => (pixel, pixel >> 8, pixel >> 16),
+                    ImageOrder::MsbFirst => (pixel >> 16, pixel >> 8, pixel),
                 };
                 data[row_start + 3 * x + 2] = p2 as u8;
                 data[row_start + 3 * x + 1] = p1 as u8;
@@ -907,8 +910,8 @@ impl<'a> Image<'a> {
             }
             BitsPerPixel::B32 => {
                 let (p0, p1, p2, p3) = match self.byte_order {
-                    ImageOrder::LSBFirst => (pixel, pixel >> 8, pixel >> 16, pixel >> 24),
-                    ImageOrder::MSBFirst => (pixel >> 24, pixel >> 16, pixel >> 8, pixel),
+                    ImageOrder::LsbFirst => (pixel, pixel >> 8, pixel >> 16, pixel >> 24),
+                    ImageOrder::MsbFirst => (pixel >> 24, pixel >> 16, pixel >> 8, pixel),
                 };
                 data[row_start + 4 * x + 3] = p3 as u8;
                 data[row_start + 4 * x + 2] = p2 as u8;
@@ -937,7 +940,7 @@ impl<'a> Image<'a> {
             BitsPerPixel::B4 => {
                 let byte = u32::from(self.data[row_start + x / 2]);
                 let odd_x = x % 2 == 1;
-                if odd_x == (self.byte_order == ImageOrder::MSBFirst) {
+                if odd_x == (self.byte_order == ImageOrder::MsbFirst) {
                     byte >> 4
                 } else {
                     byte & 0x0f
@@ -948,8 +951,8 @@ impl<'a> Image<'a> {
                 let p1 = u32::from(self.data[row_start + 2 * x + 1]);
                 let p0 = u32::from(self.data[row_start + 2 * x]);
                 match self.byte_order {
-                    ImageOrder::LSBFirst => p0 | (p1 << 8),
-                    ImageOrder::MSBFirst => p1 | (p0 << 8),
+                    ImageOrder::LsbFirst => p0 | (p1 << 8),
+                    ImageOrder::MsbFirst => p1 | (p0 << 8),
                 }
             }
             BitsPerPixel::B24 => {
@@ -957,8 +960,8 @@ impl<'a> Image<'a> {
                 let p1 = u32::from(self.data[row_start + 3 * x + 1]);
                 let p0 = u32::from(self.data[row_start + 3 * x]);
                 match self.byte_order {
-                    ImageOrder::LSBFirst => p0 | (p1 << 8) | (p2 << 16),
-                    ImageOrder::MSBFirst => p2 | (p1 << 8) | (p0 << 16),
+                    ImageOrder::LsbFirst => p0 | (p1 << 8) | (p2 << 16),
+                    ImageOrder::MsbFirst => p2 | (p1 << 8) | (p0 << 16),
                 }
             }
             BitsPerPixel::B32 => {
@@ -967,18 +970,31 @@ impl<'a> Image<'a> {
                 let p1 = u32::from(self.data[row_start + 4 * x + 1]);
                 let p0 = u32::from(self.data[row_start + 4 * x]);
                 match self.byte_order {
-                    ImageOrder::LSBFirst => p0 | (p1 << 8) | (p2 << 16) | (p3 << 24),
-                    ImageOrder::MSBFirst => p3 | (p2 << 8) | (p1 << 16) | (p0 << 24),
+                    ImageOrder::LsbFirst => p0 | (p1 << 8) | (p2 << 16) | (p3 << 24),
+                    ImageOrder::MsbFirst => p3 | (p2 << 8) | (p1 << 16) | (p0 << 24),
                 }
             }
         }
     }
+
+    /// Get a version of this image with `'static` lifetime.
+    ///
+    /// If the image was constructed from a `Cow::Borrowed`, this clones the contained data.
+    /// Otherwise, this simply returns `self`.
+    pub fn into_owned(self) -> Image<'static> {
+        // It would be great if we could just implement ToOwned, but that requires implementing
+        // Borrow, which we cannot do. Thus, this function exists as a work-around.
+        Image {
+            data: self.data.into_owned().into(),
+            ..self
+        }
+    }
 }
 
 fn compute_depth_1_address(x: usize, order: ImageOrder) -> (usize, usize) {
     let bit = match order {
-        ImageOrder::MSBFirst => 7 - x % 8,
-        ImageOrder::LSBFirst => x % 8,
+        ImageOrder::MsbFirst => 7 - x % 8,
+        ImageOrder::LsbFirst => x % 8,
     };
     (x / 8, bit)
 }
@@ -998,7 +1014,7 @@ mod test_image {
             ScanlinePad::Pad16,
             depth,
             BitsPerPixel::B8,
-            ImageOrder::MSBFirst,
+            ImageOrder::MsbFirst,
             Cow::Owned(vec![0]),
         );
         assert_eq!(result.unwrap_err(), ParseError::InsufficientData);
@@ -1013,7 +1029,7 @@ mod test_image {
             ScanlinePad::Pad16,
             depth,
             BitsPerPixel::B8,
-            ImageOrder::MSBFirst,
+            ImageOrder::MsbFirst,
             Cow::Owned(vec![42, 125]),
         )
         .unwrap();
@@ -1022,10 +1038,37 @@ mod test_image {
         assert_eq!(image.scanline_pad(), ScanlinePad::Pad16);
         assert_eq!(image.depth(), depth);
         assert_eq!(image.bits_per_pixel(), BitsPerPixel::B8);
-        assert_eq!(image.byte_order(), ImageOrder::MSBFirst);
+        assert_eq!(image.byte_order(), ImageOrder::MsbFirst);
         assert_eq!(image.data(), [42, 125]);
     }
 
+    #[test]
+    fn test_into_owned_keeps_owned_data() {
+        fn with_data(data: Cow<'_, [u8]>) -> *const u8 {
+            let orig_ptr = data.as_ptr();
+            let image = Image::new(
+                1,
+                1,
+                ScanlinePad::Pad8,
+                1,
+                BitsPerPixel::B1,
+                ImageOrder::MsbFirst,
+                data,
+            )
+            .unwrap();
+            assert_eq!(image.data().as_ptr(), orig_ptr);
+            image.into_owned().data().as_ptr()
+        }
+
+        // Cow::Borrowed is copied
+        let data = vec![0];
+        let orig_ptr = data.as_ptr();
+        assert_ne!(with_data(Cow::Borrowed(&data)), orig_ptr);
+
+        // Cow::Owned is kept
+        assert_eq!(with_data(Cow::Owned(data)), orig_ptr);
+    }
+
     #[test]
     fn put_pixel_depth1() {
         let mut image = Image::allocate(
@@ -1034,7 +1077,7 @@ mod test_image {
             ScanlinePad::Pad32,
             1,
             BitsPerPixel::B1,
-            ImageOrder::MSBFirst,
+            ImageOrder::MsbFirst,
         );
         for x in 0..8 {
             image.put_pixel(x, 0, 1);
@@ -1074,7 +1117,7 @@ mod test_image {
             ScanlinePad::Pad16,
             1,
             BitsPerPixel::B4,
-            ImageOrder::MSBFirst,
+            ImageOrder::MsbFirst,
         );
         for pos in 0..=0xf {
             image.put_pixel(pos % 8, pos / 8, pos.into());
@@ -1093,7 +1136,7 @@ mod test_image {
             ScanlinePad::Pad8,
             1,
             BitsPerPixel::B8,
-            ImageOrder::MSBFirst,
+            ImageOrder::MsbFirst,
         );
         for x in 0..=0xff {
             image.put_pixel(x, 0, x.into());
@@ -1114,7 +1157,7 @@ mod test_image {
             ScanlinePad::Pad32,
             1,
             BitsPerPixel::B16,
-            ImageOrder::MSBFirst,
+            ImageOrder::MsbFirst,
         );
         image.put_pixel(0, 0, 0xAB_36_18_F8);
         image.put_pixel(4, 0, 0x12_34_56_78);
@@ -1141,7 +1184,7 @@ mod test_image {
             ScanlinePad::Pad32,
             1,
             BitsPerPixel::B32,
-            ImageOrder::MSBFirst,
+            ImageOrder::MsbFirst,
         );
         image.put_pixel(0, 0, 0xAB_36_18_F8);
         image.put_pixel(1, 0, 0x12_34_56_78);
@@ -1164,7 +1207,7 @@ mod test_image {
             ScanlinePad::Pad32,
             1,
             BitsPerPixel::B1,
-            ImageOrder::MSBFirst,
+            ImageOrder::MsbFirst,
             Cow::Borrowed(&DATA),
         )
         .unwrap();
@@ -1184,7 +1227,7 @@ mod test_image {
             ScanlinePad::Pad32,
             1,
             BitsPerPixel::B4,
-            ImageOrder::MSBFirst,
+            ImageOrder::MsbFirst,
             Cow::Borrowed(&DATA),
         )
         .unwrap();
@@ -1204,7 +1247,7 @@ mod test_image {
             ScanlinePad::Pad32,
             1,
             BitsPerPixel::B8,
-            ImageOrder::MSBFirst,
+            ImageOrder::MsbFirst,
             Cow::Borrowed(&DATA),
         )
         .unwrap();
@@ -1224,7 +1267,7 @@ mod test_image {
             ScanlinePad::Pad32,
             1,
             BitsPerPixel::B16,
-            ImageOrder::MSBFirst,
+            ImageOrder::MsbFirst,
             Cow::Borrowed(&DATA),
         )
         .unwrap();
@@ -1244,7 +1287,7 @@ mod test_image {
             ScanlinePad::Pad32,
             1,
             BitsPerPixel::B32,
-            ImageOrder::MSBFirst,
+            ImageOrder::MsbFirst,
             Cow::Borrowed(&DATA),
         )
         .unwrap();
