@@ -8,13 +8,48 @@ use crate::errors::{ConnectionError, ParseError, ReplyError};
 use crate::protocol::xproto::{self, Atom, AtomEnum, GetPropertyReply, Window};
 use crate::x11_utils::{Serialize, TryParse};
 
+macro_rules! property_cookie {
+    {
+        $(#[$meta:meta])*
+        pub struct $cookie_name:ident: $struct_name:ident,
+        $from_reply:expr,
+    } => {
+        $(#[$meta])*
+        #[derive(Debug)]
+        pub struct $cookie_name<'a, Conn: RequestConnection + ?Sized>(Cookie<'a, Conn, GetPropertyReply>);
+
+        impl<'a, Conn> $cookie_name<'a, Conn>
+        where
+            Conn: RequestConnection + ?Sized,
+        {
+            /// Get the reply that the server sent.
+            pub fn reply(self) -> Result<Option<$struct_name>, ReplyError> {
+                #[allow(clippy::redundant_closure_call)]
+                Ok($from_reply(self.0.reply()?)?)
+            }
+
+            /// Get the reply that the server sent, but have errors handled as events.
+            pub fn reply_unchecked(self) -> Result<Option<$struct_name>, ConnectionError> {
+                self.0
+                    .reply_unchecked()?
+                    .map($from_reply)
+                    .transpose()
+                    .map(|e| e.flatten())
+                    .map_err(Into::into)
+            }
+        }
+    }
+}
+
 // WM_CLASS
 
-/// A cookie for getting a window's `WM_CLASS` property.
-///
-/// See `WmClass`.
-#[derive(Debug)]
-pub struct WmClassCookie<'a, Conn: RequestConnection + ?Sized>(Cookie<'a, Conn, GetPropertyReply>);
+property_cookie! {
+    /// A cookie for getting a window's `WM_CLASS` property.
+    ///
+    /// See `WmClass`.
+    pub struct WmClassCookie: WmClass,
+    WmClass::from_reply,
+}
 
 impl<'a, Conn> WmClassCookie<'a, Conn>
 where
@@ -32,20 +67,6 @@ where
             2048,
         )?))
     }
-
-    /// Get the reply that the server sent.
-    pub fn reply(self) -> Result<WmClass, ReplyError> {
-        Ok(WmClass::from_reply(self.0.reply()?)?)
-    }
-
-    /// Get the reply that the server sent, but have errors handled as events.
-    pub fn reply_unchecked(self) -> Result<Option<WmClass>, ConnectionError> {
-        self.0
-            .reply_unchecked()?
-            .map(WmClass::from_reply)
-            .transpose()
-            .map_err(Into::into)
-    }
 }
 
 /// The value of a window's `WM_CLASS` property.
@@ -92,7 +113,10 @@ impl WmClass {
     ///
     /// The original `GetProperty` request must have been for a `WM_CLASS` property for this
     /// function to return sensible results.
-    pub fn from_reply(reply: GetPropertyReply) -> Result<Self, ParseError> {
+    pub fn from_reply(reply: GetPropertyReply) -> Result<Option<Self>, ParseError> {
+        if reply.type_ == AtomEnum::NONE.into() {
+            return Ok(None);
+        }
         if reply.type_ != AtomEnum::STRING.into() || reply.format != 8 {
             return Err(ParseError::InvalidValue);
         }
@@ -101,8 +125,8 @@ impl WmClass {
             .value
             .iter()
             .position(|&v| v == 0)
-            .unwrap_or_else(|| reply.value.len());
-        Ok(WmClass(reply, offset))
+            .unwrap_or(reply.value.len());
+        Ok(Some(WmClass(reply, offset)))
     }
 
     /// Get the instance contained in this `WM_CLASS` property
@@ -136,13 +160,13 @@ pub enum WmSizeHintsSpecification {
     ProgramSpecified,
 }
 
-/// A cookie for getting a window's `WM_SIZE_HINTS` property.
-#[derive(Debug)]
-pub struct WmSizeHintsCookie<'a, Conn: RequestConnection + ?Sized>(
-    Cookie<'a, Conn, GetPropertyReply>,
-);
+property_cookie! {
+    /// A cookie for getting a window's `WM_SIZE_HINTS` property.
+    pub struct WmSizeHintsCookie: WmSizeHints,
+    |reply| WmSizeHints::from_reply(&reply),
+}
 
-const NUM_WM_SIZE_HINTS_ELEMENTS: u32 = 18;
+const NUM_WM_SIZE_HINTS_ELEMENTS: u16 = 18;
 
 impl<'a, Conn> WmSizeHintsCookie<'a, Conn>
 where
@@ -161,23 +185,9 @@ where
             property,
             AtomEnum::WM_SIZE_HINTS,
             0,
-            NUM_WM_SIZE_HINTS_ELEMENTS,
+            NUM_WM_SIZE_HINTS_ELEMENTS.into(),
         )?))
     }
-
-    /// Get the reply that the server sent.
-    pub fn reply(self) -> Result<WmSizeHints, ReplyError> {
-        Ok(WmSizeHints::from_reply(&self.0.reply()?)?)
-    }
-
-    /// Get the reply that the server sent, but have errors handled as events.
-    pub fn reply_unchecked(self) -> Result<Option<WmSizeHints>, ConnectionError> {
-        self.0
-            .reply_unchecked()?
-            .map(|r| WmSizeHints::from_reply(&r))
-            .transpose()
-            .map_err(Into::into)
-    }
 }
 
 // Possible flags for `WM_SIZE_HINTS`.
@@ -288,11 +298,14 @@ impl WmSizeHints {
     ///
     /// The original `WmSizeHints` request must have been for a `WM_SIZE_HINTS` property for this
     /// function to return sensible results.
-    pub fn from_reply(reply: &GetPropertyReply) -> Result<Self, ParseError> {
+    pub fn from_reply(reply: &GetPropertyReply) -> Result<Option<Self>, ParseError> {
+        if reply.type_ == AtomEnum::NONE.into() {
+            return Ok(None);
+        }
         if reply.type_ != AtomEnum::WM_SIZE_HINTS.into() || reply.format != 32 {
             return Err(ParseError::InvalidValue);
         }
-        Ok(Self::try_parse(&reply.value)?.0)
+        Ok(Some(Self::try_parse(&reply.value)?.0))
     }
 
     /// Set these `WM_SIZE_HINTS` on some window as the `WM_NORMAL_HINTS` property.
@@ -319,7 +332,7 @@ impl WmSizeHints {
             property.into(),
             AtomEnum::WM_SIZE_HINTS,
             32,
-            NUM_WM_SIZE_HINTS_ELEMENTS,
+            NUM_WM_SIZE_HINTS_ELEMENTS.into(),
             &data,
         )
     }
@@ -384,8 +397,7 @@ impl TryParse for WmSizeHints {
 impl Serialize for WmSizeHints {
     type Bytes = Vec<u8>;
     fn serialize(&self) -> Self::Bytes {
-        // 18*4 surely fits into an usize, so this unwrap() cannot trigger
-        let mut result = Vec::with_capacity((NUM_WM_SIZE_HINTS_ELEMENTS * 4).try_into().unwrap());
+        let mut result = Vec::with_capacity((NUM_WM_SIZE_HINTS_ELEMENTS * 4).into());
         self.serialize_into(&mut result);
         result
     }
@@ -433,12 +445,14 @@ impl Serialize for WmSizeHints {
 }
 
 // WM_HINTS
-
-/// A cookie for getting a window's `WM_HINTS` property.
-///
-/// See `WmHints`.
-#[derive(Debug)]
-pub struct WmHintsCookie<'a, Conn: RequestConnection + ?Sized>(Cookie<'a, Conn, GetPropertyReply>);
+//
+property_cookie! {
+    /// A cookie for getting a window's `WM_HINTS` property.
+    ///
+    /// See `WmHints`.
+    pub struct WmHintsCookie: WmHints,
+    |reply| WmHints::from_reply(&reply),
+}
 
 const NUM_WM_HINTS_ELEMENTS: u32 = 9;
 
@@ -458,20 +472,6 @@ where
             NUM_WM_HINTS_ELEMENTS,
         )?))
     }
-
-    /// Get the reply that the server sent.
-    pub fn reply(self) -> Result<WmHints, ReplyError> {
-        Ok(WmHints::from_reply(&self.0.reply()?)?)
-    }
-
-    /// Get the reply that the server sent, but have errors handled as events.
-    pub fn reply_unchecked(self) -> Result<Option<WmHints>, ConnectionError> {
-        self.0
-            .reply_unchecked()?
-            .map(|r| WmHints::from_reply(&r))
-            .transpose()
-            .map_err(Into::into)
-    }
 }
 
 /// The possible values for a `WM_STATE`'s state field.
@@ -551,12 +551,14 @@ impl WmHints {
     ///
     /// The original `WmHints` request must have been for a `WM_HINTS` property for this
     /// function to return sensible results.
-    pub fn from_reply(reply: &GetPropertyReply) -> Result<Self, ParseError> {
+    pub fn from_reply(reply: &GetPropertyReply) -> Result<Option<Self>, ParseError> {
+        if reply.type_ == AtomEnum::NONE.into() {
+            return Ok(None);
+        }
         if reply.type_ != AtomEnum::WM_HINTS.into() || reply.format != 32 {
             return Err(ParseError::InvalidValue);
         }
-
-        Ok(Self::try_parse(&reply.value)?.0)
+        Ok(Some(Self::try_parse(&reply.value)?.0))
     }
 
     /// Set these `WM_HINTS` on some window.
@@ -710,12 +712,19 @@ mod test {
             (b"Hello\0World", b"Hello", b"World"),
             (b"Hello\0World\0Good\0Day", b"Hello", b"World\0Good\0Day"),
         ] {
-            let wm_class =
-                WmClass::from_reply(get_property_reply(input, 8, AtomEnum::STRING)).unwrap();
+            let wm_class = WmClass::from_reply(get_property_reply(input, 8, AtomEnum::STRING))
+                .unwrap()
+                .unwrap();
             assert_eq!((wm_class.instance(), wm_class.class()), (*instance, *class));
         }
     }
 
+    #[test]
+    fn test_wm_class_missing() {
+        let wm_class = WmClass::from_reply(get_property_reply(&[], 0, AtomEnum::NONE)).unwrap();
+        assert!(wm_class.is_none());
+    }
+
     #[test]
     fn test_wm_normal_hints() {
         // This is the value of some random xterm window.
@@ -746,6 +755,7 @@ mod test {
             .collect::<Vec<u8>>();
         let wm_size_hints =
             WmSizeHints::from_reply(&get_property_reply(&input, 32, AtomEnum::WM_SIZE_HINTS))
+                .unwrap()
                 .unwrap();
 
         assert!(
@@ -764,6 +774,13 @@ mod test {
         assert_eq!(input, wm_size_hints.serialize());
     }
 
+    #[test]
+    fn test_wm_normal_hints_missing() {
+        let wm_size_hints =
+            WmSizeHints::from_reply(&get_property_reply(&[], 0, AtomEnum::NONE)).unwrap();
+        assert!(wm_size_hints.is_none());
+    }
+
     #[test]
     fn test_wm_hints() {
         // This is the value of some random xterm window.
@@ -783,8 +800,9 @@ mod test {
             .iter()
             .flat_map(|v| u32::serialize(v).to_vec())
             .collect::<Vec<u8>>();
-        let wm_hints =
-            WmHints::from_reply(&get_property_reply(&input, 32, AtomEnum::WM_HINTS)).unwrap();
+        let wm_hints = WmHints::from_reply(&get_property_reply(&input, 32, AtomEnum::WM_HINTS))
+            .unwrap()
+            .unwrap();
 
         assert_eq!(wm_hints.input, Some(true));
         match wm_hints.initial_state {
@@ -796,8 +814,14 @@ mod test {
         assert_eq!(wm_hints.icon_position, None);
         assert_eq!(wm_hints.icon_mask, None);
         assert_eq!(wm_hints.window_group, Some(0x0060_0009));
-        assert_eq!(wm_hints.urgent, false);
+        assert!(!wm_hints.urgent);
 
         assert_eq!(input, wm_hints.serialize());
     }
+
+    #[test]
+    fn test_wm_hints_missing() {
+        let wm_hints = WmHints::from_reply(&get_property_reply(&[], 0, AtomEnum::NONE)).unwrap();
+        assert!(wm_hints.is_none());
+    }
 }
