@@ -2,36 +2,35 @@
 
 use std::convert::TryInto;
 use std::io::IoSlice;
+use std::mem::drop;
 use std::sync::{Condvar, Mutex, MutexGuard, TryLockError};
+use std::time::Instant;
 
 use crate::connection::{
-    compute_length_field, Connection, DiscardMode, ReplyOrError, RequestConnection, RequestKind,
-    SequenceNumber,
+    compute_length_field, Connection, ReplyOrError, RequestConnection, RequestKind,
 };
 use crate::cookie::{Cookie, CookieWithFds, VoidCookie};
+use crate::errors::DisplayParsingError;
 pub use crate::errors::{ConnectError, ConnectionError, ParseError, ReplyError, ReplyOrIdError};
 use crate::extension_manager::ExtensionManager;
 use crate::protocol::bigreq::{ConnectionExt as _, EnableReply};
-use crate::protocol::xproto::{Setup, SetupRequest, GET_INPUT_FOCUS_REQUEST};
+use crate::protocol::xproto::{Setup, GET_INPUT_FOCUS_REQUEST, QUERY_EXTENSION_REQUEST};
 use crate::utils::RawFdContainer;
-use crate::x11_utils::{ExtensionInformation, Serialize, TryParse, TryParseFd};
+use crate::x11_utils::{ExtensionInformation, TryParse, TryParseFd};
+use x11rb_protocol::connect::Connect;
+use x11rb_protocol::connection::{Connection as ProtoConnection, PollReply, ReplyFdKind};
+use x11rb_protocol::id_allocator::IdAllocator;
+use x11rb_protocol::{xauth::get_auth, DiscardMode, RawEventAndSeqNumber, SequenceNumber};
 
-mod id_allocator;
-mod inner;
 mod packet_reader;
-mod parse_display;
 mod stream;
 mod write_buffer;
-mod xauth;
 
-use inner::PollReply;
 use packet_reader::PacketReader;
 pub use stream::{DefaultStream, PollMode, Stream};
 use write_buffer::WriteBuffer;
 
 type Buffer = <RustConnection as RequestConnection>::Buf;
-/// The raw bytes of an event received by [`RustConnection`] and its sequence number.
-pub type RawEventAndSeqNumber = crate::connection::RawEventAndSeqNumber<Buffer>;
 /// A combination of a buffer and a list of file descriptors for use by [`RustConnection`].
 pub type BufWithFds = crate::connection::BufWithFds<Buffer>;
 
@@ -42,15 +41,14 @@ enum MaxRequestBytes {
     Known(usize),
 }
 
-type MutexGuardInner<'a> = MutexGuard<'a, inner::ConnectionInner>;
-
-#[derive(Debug, Copy, Clone, PartialEq, Eq)]
-pub(crate) enum ReplyFDKind {
-    NoReply,
-    ReplyWithoutFDs,
-    ReplyWithFDs,
+#[derive(Debug)]
+struct ConnectionInner {
+    inner: ProtoConnection,
+    write_buffer: WriteBuffer,
 }
 
+type MutexGuardInner<'a> = MutexGuard<'a, ConnectionInner>;
+
 #[derive(Debug, Copy, Clone, PartialEq, Eq)]
 pub(crate) enum BlockingMode {
     Blocking,
@@ -67,16 +65,16 @@ pub(crate) enum BlockingMode {
 /// to be buffered.
 #[derive(Debug)]
 pub struct RustConnection<S: Stream = DefaultStream> {
-    inner: Mutex<inner::ConnectionInner>,
+    inner: Mutex<ConnectionInner>,
     stream: S,
     // This mutex is only locked with `try_lock` (never blocks), so a simpler
     // lock based only on a atomic variable would be more efficient.
     packet_reader: Mutex<PacketReader>,
     reader_condition: Condvar,
-    id_allocator: Mutex<id_allocator::IDAllocator>,
     setup: Setup,
     extension_manager: Mutex<ExtensionManager>,
     maximum_request_bytes: Mutex<MaxRequestBytes>,
+    id_allocator: Mutex<IdAllocator>,
 }
 
 // Locking rules
@@ -103,6 +101,8 @@ pub struct RustConnection<S: Stream = DefaultStream> {
 // read while `read` is already locked, they sleep on `reader_condition`. The actual reader will
 // then notify this condition variable once it is done reading.
 //
+// n.b. notgull: write_buffer follows the same rules
+//
 // The condition variable is necessary since one thread may read packets that another thread waits
 // for. Thus, after reading something from the connection, all threads that wait for something have
 // to check if they are the intended recipient.
@@ -113,25 +113,50 @@ impl RustConnection<DefaultStream> {
     /// If no `dpy_name` is provided, the value from `$DISPLAY` is used.
     pub fn connect(dpy_name: Option<&str>) -> Result<(Self, usize), ConnectError> {
         // Parse display information
-        let parsed_display =
-            parse_display::parse_display(dpy_name).ok_or(ConnectError::DisplayParsingError)?;
-
-        // Establish connection
-        let protocol = parsed_display.protocol.as_deref();
-        let stream =
-            DefaultStream::connect(&*parsed_display.host, protocol, parsed_display.display)?;
+        let parsed_display = x11rb_protocol::parse_display::parse_display(dpy_name)?;
         let screen = parsed_display.screen.into();
 
-        let (family, address) = stream.peer_addr()?;
-        let (auth_name, auth_data) = xauth::get_auth(family, &address, parsed_display.display)
-            // Ignore all errors while determining auth; instead we just try without auth info.
-            .unwrap_or(None)
-            .unwrap_or_else(|| (Vec::new(), Vec::new()));
+        // Establish connection by iterating over ConnectAddresses until we find one that
+        // works.
+        let mut error = None;
+        for addr in parsed_display.connect_instruction() {
+            let start = Instant::now();
+            match DefaultStream::connect(&addr) {
+                Ok((stream, (family, address))) => {
+                    crate::trace!(
+                        "Connected to X11 server via {:?} in {:?}",
+                        addr,
+                        start.elapsed()
+                    );
 
-        Ok((
-            Self::connect_to_stream_with_auth_info(stream, screen, auth_name, auth_data)?,
-            screen,
-        ))
+                    // we found a stream, get auth information
+                    let (auth_name, auth_data) = get_auth(family, &address, parsed_display.display)
+                        // Ignore all errors while determining auth; instead we just try without auth info.
+                        .unwrap_or(None)
+                        .unwrap_or_else(|| (Vec::new(), Vec::new()));
+                    crate::trace!("Picked authentication via auth mechanism {:?}", auth_name);
+
+                    // finish connecting to server
+                    return Ok((
+                        Self::connect_to_stream_with_auth_info(
+                            stream, screen, auth_name, auth_data,
+                        )?,
+                        screen,
+                    ));
+                }
+                Err(e) => {
+                    crate::debug!("Failed to connect to X11 server via {:?}: {:?}", addr, e);
+                    error = Some(e);
+                    continue;
+                }
+            }
+        }
+
+        // none of the addresses worked
+        Err(match error {
+            Some(e) => ConnectError::IoError(e),
+            None => DisplayParsingError::Unknown.into(),
+        })
     }
 }
 
@@ -160,8 +185,64 @@ impl<S: Stream> RustConnection<S> {
         auth_name: Vec<u8>,
         auth_data: Vec<u8>,
     ) -> Result<Self, ConnectError> {
-        write_setup(&stream, auth_name, auth_data)?;
-        let setup = read_setup(&stream)?;
+        let (mut connect, setup_request) = Connect::with_authorization(auth_name, auth_data);
+
+        // write the connect() setup request
+        let mut nwritten = 0;
+        let mut fds = vec![];
+
+        crate::trace!(
+            "Writing connection setup with {} bytes",
+            setup_request.len()
+        );
+        while nwritten != setup_request.len() {
+            stream.poll(PollMode::Writable)?;
+            // poll returned successfully, so the stream is writable.
+            match stream.write(&setup_request[nwritten..], &mut fds) {
+                Ok(0) => {
+                    return Err(std::io::Error::new(
+                        std::io::ErrorKind::WriteZero,
+                        "failed to write whole buffer",
+                    )
+                    .into())
+                }
+                Ok(n) => nwritten += n,
+                // Spurious wakeup from poll, try again
+                Err(ref e) if e.kind() == std::io::ErrorKind::WouldBlock => {}
+                Err(e) => return Err(e.into()),
+            }
+        }
+
+        // read in the setup
+        loop {
+            stream.poll(PollMode::Readable)?;
+            crate::trace!(
+                "Reading connection setup with at least {} bytes remaining",
+                connect.buffer().len()
+            );
+            let adv = match stream.read(connect.buffer(), &mut fds) {
+                Ok(0) => {
+                    return Err(std::io::Error::new(
+                        std::io::ErrorKind::UnexpectedEof,
+                        "failed to read whole buffer",
+                    )
+                    .into())
+                }
+                Ok(n) => n,
+                // Spurious wakeup from poll, try again
+                Err(ref e) if e.kind() == std::io::ErrorKind::WouldBlock => continue,
+                Err(e) => return Err(e.into()),
+            };
+            crate::trace!("Read {} bytes", adv);
+
+            // advance the internal buffer
+            if connect.advance(adv) {
+                break;
+            }
+        }
+
+        // resolve the setup
+        let setup = connect.into_setup()?;
 
         // Check that we got a valid screen number
         if screen >= setup.roots.len() {
@@ -174,29 +255,24 @@ impl<S: Stream> RustConnection<S> {
 
     /// Establish a new connection for an already connected stream.
     ///
-    /// `read` is used for reading data from the X11 server and `write` is used for writing.
+    /// The given `stream` is used for communicating with the X11 server.
     /// It is assumed that `setup` was just received from the server. Thus, the first reply to a
     /// request that is sent will have sequence number one.
     pub fn for_connected_stream(stream: S, setup: Setup) -> Result<Self, ConnectError> {
-        Self::for_inner(stream, inner::ConnectionInner::new(), setup)
-    }
+        let id_allocator = IdAllocator::new(setup.resource_id_base, setup.resource_id_mask)?;
 
-    fn for_inner(
-        stream: S,
-        inner: inner::ConnectionInner,
-        setup: Setup,
-    ) -> Result<Self, ConnectError> {
-        let allocator =
-            id_allocator::IDAllocator::new(setup.resource_id_base, setup.resource_id_mask)?;
         Ok(RustConnection {
-            inner: Mutex::new(inner),
+            inner: Mutex::new(ConnectionInner {
+                inner: ProtoConnection::new(),
+                write_buffer: WriteBuffer::new(),
+            }),
             stream,
             packet_reader: Mutex::new(PacketReader::new()),
             reader_condition: Condvar::new(),
-            id_allocator: Mutex::new(allocator),
             setup,
             extension_manager: Default::default(),
             maximum_request_bytes: Mutex::new(MaxRequestBytes::Unknown),
+            id_allocator: Mutex::new(id_allocator),
         })
     }
 
@@ -208,25 +284,36 @@ impl<S: Stream> RustConnection<S> {
         &self,
         bufs: &[IoSlice<'_>],
         fds: Vec<RawFdContainer>,
-        kind: ReplyFDKind,
+        kind: ReplyFdKind,
     ) -> Result<SequenceNumber, ConnectionError> {
+        let _guard = crate::debug_span!("send_request").entered();
+
+        let request_info = RequestInfo {
+            extension_manager: &self.extension_manager,
+            major_opcode: bufs[0][0],
+            minor_opcode: bufs[0][1],
+        };
+        crate::debug!("Sending {}", request_info);
+
         let mut storage = Default::default();
         let bufs = compute_length_field(self, bufs, &mut storage)?;
 
         // Note: `inner` must be kept blocked until the request has been completely written
         // or buffered to avoid sending the data of different requests interleaved. For this
         // reason, `read_packet_and_enqueue` must always be called with `BlockingMode::NonBlocking`
-        // during a write, otherise `inner` would be temporarily released.
+        // during a write, otherwise `inner` would be temporarily released.
         let mut inner = self.inner.lock().unwrap();
 
         loop {
-            match inner.send_request(kind) {
+            let send_result = inner.inner.send_request(kind);
+            match send_result {
                 Some(seqno) => {
                     // Now actually send the buffers
                     let _inner = self.write_all_vectored(inner, bufs, fds)?;
                     return Ok(seqno);
                 }
                 None => {
+                    crate::trace!("Syncing with the X11 server since there are too many outstanding void requests");
                     inner = self.send_sync(inner)?;
                 }
             }
@@ -251,9 +338,12 @@ impl<S: Stream> RustConnection<S> {
         ];
 
         let seqno = inner
-            .send_request(ReplyFDKind::ReplyWithoutFDs)
+            .inner
+            .send_request(ReplyFdKind::ReplyWithoutFDs)
             .expect("Sending a HasResponse request should not be blocked by syncs");
-        inner.discard_reply(seqno, DiscardMode::DiscardReplyAndError);
+        inner
+            .inner
+            .discard_reply(seqno, DiscardMode::DiscardReplyAndError);
         let inner = self.write_all_vectored(inner, &[IoSlice::new(&request)], Vec::new())?;
 
         Ok(inner)
@@ -268,13 +358,15 @@ impl<S: Stream> RustConnection<S> {
         mut fds: Vec<RawFdContainer>,
     ) -> std::io::Result<MutexGuardInner<'a>> {
         let mut partial_buf: &[u8] = &[];
-        while !partial_buf.is_empty() || !bufs.is_empty() || !fds.is_empty() {
+        while !partial_buf.is_empty() || !bufs.is_empty() {
             self.stream.poll(PollMode::ReadAndWritable)?;
             let write_result = if !partial_buf.is_empty() {
+                // "inner" is held, passed into this function, so this should never be held
                 inner
                     .write_buffer
                     .write(&self.stream, partial_buf, &mut fds)
             } else {
+                // same as above
                 inner
                     .write_buffer
                     .write_vectored(&self.stream, bufs, &mut fds)
@@ -310,6 +402,7 @@ impl<S: Stream> RustConnection<S> {
                     }
                 }
                 Err(ref e) if e.kind() == std::io::ErrorKind::WouldBlock => {
+                    crate::trace!("Writing more data would block for now");
                     // Writing would block, try to read instead because the
                     // server might not accept new requests after its
                     // buffered replies have been read.
@@ -318,6 +411,12 @@ impl<S: Stream> RustConnection<S> {
                 Err(e) => return Err(e),
             }
         }
+        if !fds.is_empty() {
+            return Err(std::io::Error::new(
+                std::io::ErrorKind::Other,
+                "Left over FDs after sending the request",
+            ));
+        }
         Ok(inner)
     }
 
@@ -325,12 +424,15 @@ impl<S: Stream> RustConnection<S> {
         &'a self,
         mut inner: MutexGuardInner<'a>,
     ) -> std::io::Result<MutexGuardInner<'a>> {
+        // n.b. notgull: inner guard is held
         while inner.write_buffer.needs_flush() {
             self.stream.poll(PollMode::ReadAndWritable)?;
-            match inner.write_buffer.flush(&self.stream) {
+            let flush_result = inner.write_buffer.flush(&self.stream);
+            match flush_result {
                 // Flush completed
                 Ok(()) => break,
                 Err(ref e) if e.kind() == std::io::ErrorKind::WouldBlock => {
+                    crate::trace!("Flushing more data would block for now");
                     // Writing would block, try to read instead because the
                     // server might not accept new requests after its
                     // buffered replies have been read.
@@ -363,8 +465,13 @@ impl<S: Stream> RustConnection<S> {
             Err(TryLockError::WouldBlock) => {
                 // In non-blocking mode, we just return immediately
                 match mode {
-                    BlockingMode::NonBlocking => return Ok(inner),
-                    BlockingMode::Blocking => {}
+                    BlockingMode::NonBlocking => {
+                        crate::trace!("read_packet_and_enqueue in NonBlocking mode doing nothing since reader is already locked");
+                        return Ok(inner);
+                    }
+                    BlockingMode::Blocking => {
+                        crate::trace!("read_packet_and_enqueue in Blocking mode waiting for pre-existing reader");
+                    }
                 }
 
                 // 1.1. Someone else is reading (other thread is at 2.2);
@@ -409,10 +516,10 @@ impl<S: Stream> RustConnection<S> {
                 drop(packet_reader);
 
                 // 2.4. Actually enqueue the read packets.
-                inner.enqueue_fds(fds);
+                inner.inner.enqueue_fds(fds);
                 packets
                     .into_iter()
-                    .for_each(|packet| inner.enqueue_packet(packet));
+                    .for_each(|packet| inner.inner.enqueue_packet(packet));
 
                 // 2.5. Notify the condvar by dropping the `notify_on_drop` object.
                 // The object would have been dropped when the function returns, so
@@ -428,6 +535,7 @@ impl<S: Stream> RustConnection<S> {
 
     fn prefetch_maximum_request_bytes_impl(&self, max_bytes: &mut MutexGuard<'_, MaxRequestBytes>) {
         if let MaxRequestBytes::Unknown = **max_bytes {
+            crate::info!("Prefetching maximum request length");
             let request = self
                 .bigreq_enable()
                 .map(|cookie| cookie.into_sequence_number())
@@ -455,7 +563,7 @@ impl<S: Stream> RequestConnection for RustConnection<S> {
     {
         Ok(Cookie::new(
             self,
-            self.send_request(bufs, fds, ReplyFDKind::ReplyWithoutFDs)?,
+            self.send_request(bufs, fds, ReplyFdKind::ReplyWithoutFDs)?,
         ))
     }
 
@@ -469,7 +577,7 @@ impl<S: Stream> RequestConnection for RustConnection<S> {
     {
         Ok(CookieWithFds::new(
             self,
-            self.send_request(bufs, fds, ReplyFDKind::ReplyWithFDs)?,
+            self.send_request(bufs, fds, ReplyFdKind::ReplyWithFDs)?,
         ))
     }
 
@@ -480,12 +588,21 @@ impl<S: Stream> RequestConnection for RustConnection<S> {
     ) -> Result<VoidCookie<'_, Self>, ConnectionError> {
         Ok(VoidCookie::new(
             self,
-            self.send_request(bufs, fds, ReplyFDKind::NoReply)?,
+            self.send_request(bufs, fds, ReplyFdKind::NoReply)?,
         ))
     }
 
     fn discard_reply(&self, sequence: SequenceNumber, _kind: RequestKind, mode: DiscardMode) {
-        self.inner.lock().unwrap().discard_reply(sequence, mode);
+        crate::debug!(
+            "Discarding reply to request {} in mode {:?}",
+            sequence,
+            mode
+        );
+        self.inner
+            .lock()
+            .unwrap()
+            .inner
+            .discard_reply(sequence, mode);
     }
 
     fn prefetch_extension_information(
@@ -519,10 +636,14 @@ impl<S: Stream> RequestConnection for RustConnection<S> {
     }
 
     fn wait_for_reply(&self, sequence: SequenceNumber) -> Result<Option<Vec<u8>>, ConnectionError> {
+        let _guard = crate::debug_span!("wait_for_reply", sequence).entered();
+
         let mut inner = self.inner.lock().unwrap();
         inner = self.flush_impl(inner)?;
         loop {
-            match inner.poll_for_reply(sequence) {
+            crate::trace!({ sequence }, "Polling for reply");
+            let poll_result = inner.inner.poll_for_reply(sequence);
+            match poll_result {
                 PollReply::TryAgain => {}
                 PollReply::NoReply => return Ok(None),
                 PollReply::Reply(buffer) => return Ok(Some(buffer)),
@@ -535,15 +656,20 @@ impl<S: Stream> RequestConnection for RustConnection<S> {
         &self,
         sequence: SequenceNumber,
     ) -> Result<Option<Buffer>, ConnectionError> {
+        let _guard = crate::debug_span!("check_for_raw_error", sequence).entered();
+
         let mut inner = self.inner.lock().unwrap();
-        if inner.prepare_check_for_reply_or_error(sequence) {
+        if inner.inner.prepare_check_for_reply_or_error(sequence) {
+            crate::trace!("Inserting sync with the X11 server");
             inner = self.send_sync(inner)?;
-            assert!(!inner.prepare_check_for_reply_or_error(sequence));
+            assert!(!inner.inner.prepare_check_for_reply_or_error(sequence));
         }
         // Ensure the request is sent
         inner = self.flush_impl(inner)?;
         loop {
-            match inner.poll_check_for_reply_or_error(sequence) {
+            crate::trace!({ sequence }, "Polling for reply or error");
+            let poll_result = inner.inner.poll_check_for_reply_or_error(sequence);
+            match poll_result {
                 PollReply::TryAgain => {}
                 PollReply::NoReply => return Ok(None),
                 PollReply::Reply(buffer) => return Ok(Some(buffer)),
@@ -556,14 +682,19 @@ impl<S: Stream> RequestConnection for RustConnection<S> {
         &self,
         sequence: SequenceNumber,
     ) -> Result<ReplyOrError<BufWithFds, Buffer>, ConnectionError> {
+        let _guard = crate::debug_span!("wait_for_reply_with_fds_raw", sequence).entered();
+
         let mut inner = self.inner.lock().unwrap();
         // Ensure the request is sent
         inner = self.flush_impl(inner)?;
         loop {
-            if let Some(reply) = inner.poll_for_reply_or_error(sequence) {
+            crate::trace!({ sequence }, "Polling for reply or error");
+            if let Some(reply) = inner.inner.poll_for_reply_or_error(sequence) {
                 if reply.0[0] == 0 {
+                    crate::trace!("Got error");
                     return Ok(ReplyOrError::Error(reply.0));
                 } else {
+                    crate::trace!("Got reply");
                     return Ok(ReplyOrError::Reply(reply));
                 }
             }
@@ -575,9 +706,12 @@ impl<S: Stream> RequestConnection for RustConnection<S> {
         let mut max_bytes = self.maximum_request_bytes.lock().unwrap();
         self.prefetch_maximum_request_bytes_impl(&mut max_bytes);
         use MaxRequestBytes::*;
-        match *max_bytes {
+        let max_bytes = &mut *max_bytes;
+        match max_bytes {
             Unknown => unreachable!("We just prefetched this"),
             Requested(seqno) => {
+                let _guard = crate::info_span!("maximum_request_bytes").entered();
+
                 let length = seqno
                     // If prefetching the request succeeded, get a cookie
                     .and_then(|seqno| {
@@ -594,9 +728,10 @@ impl<S: Stream> RequestConnection for RustConnection<S> {
                     .unwrap_or(usize::max_value());
                 let length = length * 4;
                 *max_bytes = Known(length);
+                crate::info!("Maximum request length is {} bytes", length);
                 length
             }
-            Known(length) => length,
+            Known(length) => *length,
         }
     }
 
@@ -617,10 +752,14 @@ impl<S: Stream> RequestConnection for RustConnection<S> {
 }
 
 impl<S: Stream> Connection for RustConnection<S> {
-    fn wait_for_raw_event_with_sequence(&self) -> Result<RawEventAndSeqNumber, ConnectionError> {
+    fn wait_for_raw_event_with_sequence(
+        &self,
+    ) -> Result<RawEventAndSeqNumber<Vec<u8>>, ConnectionError> {
+        let _guard = crate::trace_span!("wait_for_raw_event_with_sequence").entered();
+
         let mut inner = self.inner.lock().unwrap();
         loop {
-            if let Some(event) = inner.poll_for_event_with_sequence() {
+            if let Some(event) = inner.inner.poll_for_event_with_sequence() {
                 return Ok(event);
             }
             inner = self.read_packet_and_enqueue(inner, BlockingMode::Blocking)?;
@@ -629,13 +768,15 @@ impl<S: Stream> Connection for RustConnection<S> {
 
     fn poll_for_raw_event_with_sequence(
         &self,
-    ) -> Result<Option<RawEventAndSeqNumber>, ConnectionError> {
+    ) -> Result<Option<RawEventAndSeqNumber<Vec<u8>>>, ConnectionError> {
+        let _guard = crate::trace_span!("poll_for_raw_event_with_sequence").entered();
+
         let mut inner = self.inner.lock().unwrap();
-        if let Some(event) = inner.poll_for_event_with_sequence() {
+        if let Some(event) = inner.inner.poll_for_event_with_sequence() {
             Ok(Some(event))
         } else {
             inner = self.read_packet_and_enqueue(inner, BlockingMode::NonBlocking)?;
-            Ok(inner.poll_for_event_with_sequence())
+            Ok(inner.inner.poll_for_event_with_sequence())
         }
     }
 
@@ -650,90 +791,27 @@ impl<S: Stream> Connection for RustConnection<S> {
     }
 
     fn generate_id(&self) -> Result<u32, ReplyOrIdError> {
-        self.id_allocator.lock().unwrap().generate_id(self)
-    }
-}
+        let mut id_allocator = self.id_allocator.lock().unwrap();
+        if let Some(id) = id_allocator.generate_id() {
+            Ok(id)
+        } else {
+            use crate::protocol::xc_misc::{self, ConnectionExt as _};
 
-#[cfg(target_endian = "little")]
-fn byte_order() -> u8 {
-    0x6c
-}
-
-#[cfg(target_endian = "big")]
-fn byte_order() -> u8 {
-    0x42
-}
-
-/// Send a `SetupRequest` to the X11 server.
-fn write_setup(
-    write: &impl Stream,
-    auth_name: Vec<u8>,
-    auth_data: Vec<u8>,
-) -> Result<(), std::io::Error> {
-    let request = SetupRequest {
-        byte_order: byte_order(),
-        protocol_major_version: 11,
-        protocol_minor_version: 0,
-        authorization_protocol_name: auth_name,
-        authorization_protocol_data: auth_data,
-    };
-    // Write the data
-    let data = request.serialize();
-    let mut nwritten = 0;
-    while nwritten != data.len() {
-        write.poll(PollMode::Writable)?;
-        // poll returned successfully, so the stream is writable.
-        match write.write(&data[nwritten..], &mut Vec::new()) {
-            Ok(0) => {
-                return Err(std::io::Error::new(
-                    std::io::ErrorKind::WriteZero,
-                    "failed to write whole buffer",
-                ))
+            if self
+                .extension_information(xc_misc::X11_EXTENSION_NAME)?
+                .is_none()
+            {
+                crate::error!("XIDs are exhausted and XC-MISC extension is not available");
+                Err(ReplyOrIdError::IdsExhausted)
+            } else {
+                crate::info!("XIDs are exhausted; fetching free range via XC-MISC");
+                id_allocator.update_xid_range(&self.xc_misc_get_xid_range()?.reply()?)?;
+                id_allocator
+                    .generate_id()
+                    .ok_or(ReplyOrIdError::IdsExhausted)
             }
-            Ok(n) => nwritten += n,
-            // Spurious wakeup from poll, try again
-            Err(ref e) if e.kind() == std::io::ErrorKind::WouldBlock => {}
-            Err(e) => return Err(e),
         }
     }
-    Ok(())
-}
-
-/// Read a `Setup` from the X11 server.
-///
-/// If the server sends a `SetupFailed` or `SetupAuthenticate` packet, these will be returned
-/// as errors.
-fn read_setup(stream: &impl Stream) -> Result<Setup, ConnectError> {
-    let mut fds = Vec::new();
-    let mut setup = vec![0; 8];
-    stream.read_exact(&mut setup, &mut fds)?;
-    let extra_length = usize::from(u16::from_ne_bytes([setup[6], setup[7]])) * 4;
-    // Use `Vec::reserve_exact` because this will be the final
-    // length of the vector.
-    setup.reserve_exact(extra_length);
-    setup.resize(8 + extra_length, 0);
-    stream.read_exact(&mut setup[8..], &mut fds)?;
-    if !fds.is_empty() {
-        return Err(std::io::Error::new(
-            std::io::ErrorKind::Other,
-            "unexpectedly received FDs in connection setup",
-        )
-        .into());
-    }
-    match setup[0] {
-        // 0 is SetupFailed
-        0 => Err(ConnectError::SetupFailed(
-            TryParse::try_parse(&setup[..])?.0,
-        )),
-        // Success
-        1 => Ok(Setup::try_parse(&setup[..])?.0),
-        // 2 is SetupAuthenticate
-        2 => Err(ConnectError::SetupAuthenticate(
-            TryParse::try_parse(&setup[..])?.0,
-        )),
-        // Uhm... no other cases are defined
-        _ => Err(ParseError::InvalidValue.into()),
-    }
 }
 
 /// Call `notify_all` on a condition variable when dropped.
@@ -746,107 +824,30 @@ impl Drop for NotifyOnDrop<'_> {
     }
 }
 
-#[cfg(test)]
-mod test {
-    use std::cell::RefCell;
-    use std::io::{Read, Result, Write};
+/// Format information about a request in a Display impl
+struct RequestInfo<'a> {
+    extension_manager: &'a Mutex<ExtensionManager>,
+    major_opcode: u8,
+    minor_opcode: u8,
+}
 
-    use super::{read_setup, PollMode, Stream};
-    use crate::errors::ConnectError;
-    use crate::protocol::xproto::{ImageOrder, Setup, SetupAuthenticate, SetupFailed};
-    use crate::utils::RawFdContainer;
-    use crate::x11_utils::Serialize;
-
-    struct SliceStream<'a, 'b> {
-        read_slice: RefCell<&'a [u8]>,
-        write_slice: RefCell<&'b mut [u8]>,
-    }
-
-    impl<'a, 'b> Stream for SliceStream<'a, 'b> {
-        fn poll(&self, _mode: PollMode) -> Result<()> {
-            Ok(())
-        }
-
-        fn read(&self, buf: &mut [u8], _fd_storage: &mut Vec<RawFdContainer>) -> Result<usize> {
-            self.read_slice.borrow_mut().read(buf)
-        }
-
-        fn write(&self, buf: &[u8], fds: &mut Vec<RawFdContainer>) -> Result<usize> {
-            assert!(fds.is_empty());
-            self.write_slice.borrow_mut().write(buf)
-        }
-    }
-
-    #[test]
-    fn read_setup_success() {
-        let mut setup = Setup {
-            status: 1,
-            protocol_major_version: 11,
-            protocol_minor_version: 0,
-            length: 0,
-            release_number: 0,
-            resource_id_base: 0,
-            resource_id_mask: 0,
-            motion_buffer_size: 0,
-            maximum_request_length: 0,
-            image_byte_order: ImageOrder::LSB_FIRST,
-            bitmap_format_bit_order: ImageOrder::LSB_FIRST,
-            bitmap_format_scanline_unit: 0,
-            bitmap_format_scanline_pad: 0,
-            min_keycode: 0,
-            max_keycode: 0,
-            vendor: vec![],
-            pixmap_formats: vec![],
-            roots: vec![],
-        };
-        setup.length = ((setup.serialize().len() - 8) / 4) as _;
-        let setup_bytes = setup.serialize();
-
-        let stream = SliceStream {
-            read_slice: RefCell::new(&setup_bytes),
-            write_slice: RefCell::new(&mut []),
-        };
-        let read = read_setup(&stream);
-        assert_eq!(setup, read.unwrap());
-    }
-
-    #[test]
-    fn read_setup_failed() {
-        let mut setup = SetupFailed {
-            status: 0,
-            protocol_major_version: 11,
-            protocol_minor_version: 0,
-            length: 0,
-            reason: b"whatever".to_vec(),
-        };
-        setup.length = ((setup.serialize().len() - 8) / 4) as _;
-        let setup_bytes = setup.serialize();
-
-        let stream = SliceStream {
-            read_slice: RefCell::new(&setup_bytes),
-            write_slice: RefCell::new(&mut []),
-        };
-        match read_setup(&stream) {
-            Err(ConnectError::SetupFailed(read)) => assert_eq!(setup, read),
-            value => panic!("Unexpected value {:?}", value),
-        }
-    }
-
-    #[test]
-    fn read_setup_authenticate() {
-        let setup = SetupAuthenticate {
-            status: 2,
-            reason: b"12345678".to_vec(),
-        };
-        let setup_bytes = setup.serialize();
-
-        let stream = SliceStream {
-            read_slice: RefCell::new(&setup_bytes),
-            write_slice: RefCell::new(&mut []),
-        };
-        match read_setup(&stream) {
-            Err(ConnectError::SetupAuthenticate(read)) => assert_eq!(setup, read),
-            value => panic!("Unexpected value {:?}", value),
+impl std::fmt::Display for RequestInfo<'_> {
+    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
+        // QueryExtension is used by the extension manager. We would deadlock if we
+        // tried to lock it again. Hence, this case is hardcoded here.
+        if self.major_opcode == QUERY_EXTENSION_REQUEST {
+            write!(f, "QueryExtension request")
+        } else {
+            let guard = self.extension_manager.lock().unwrap();
+            write!(
+                f,
+                "{} request",
+                x11rb_protocol::protocol::get_request_name(
+                    &*guard,
+                    self.major_opcode,
+                    self.minor_opcode
+                )
+            )
         }
     }
 }
