@@ -7,35 +7,36 @@ use std::ffi::CStr;
 use std::io::{Error as IOError, ErrorKind, IoSlice};
 use std::os::raw::c_int;
 #[cfg(unix)]
-use std::os::unix::io::{AsRawFd, RawFd};
+use std::os::unix::io::{AsFd, AsRawFd, BorrowedFd, FromRawFd, IntoRawFd, OwnedFd, RawFd};
 use std::ptr::{null, null_mut};
-use std::sync::{
-    atomic::{AtomicU64, Ordering},
-    Mutex,
-};
+use std::sync::{atomic::Ordering, Mutex};
 
 use libc::c_void;
 
 use crate::connection::{
-    compute_length_field, Connection, DiscardMode, ReplyOrError, RequestConnection, RequestKind,
-    SequenceNumber,
+    compute_length_field, Connection, ReplyOrError, RequestConnection, RequestKind,
 };
 use crate::cookie::{Cookie, CookieWithFds, VoidCookie};
+use crate::errors::DisplayParsingError;
 pub use crate::errors::{ConnectError, ConnectionError, ParseError, ReplyError, ReplyOrIdError};
 use crate::extension_manager::ExtensionManager;
 use crate::protocol::xproto::Setup;
 use crate::utils::{CSlice, RawFdContainer};
 use crate::x11_utils::{ExtensionInformation, TryParse, TryParseFd};
 
+use x11rb_protocol::{DiscardMode, SequenceNumber};
+
+mod atomic_u64;
 mod pending_errors;
 mod raw_ffi;
 
+use atomic_u64::AtomicU64;
 #[cfg(all(not(test), feature = "dl-libxcb"))]
 pub use raw_ffi::libxcb_library::load_libxcb;
 
 type Buffer = <XCBConnection as RequestConnection>::Buf;
 /// The raw bytes of an event received by [`XCBConnection`] and its sequence number.
-pub type RawEventAndSeqNumber = crate::connection::RawEventAndSeqNumber<Buffer>;
+pub type RawEventAndSeqNumber = x11rb_protocol::RawEventAndSeqNumber<Buffer>;
 /// A combination of a buffer and a list of file descriptors for use by [`XCBConnection`].
 pub type BufWithFds = crate::connection::BufWithFds<Buffer>;
 
@@ -43,9 +44,10 @@ pub type BufWithFds = crate::connection::BufWithFds<Buffer>;
 ///
 /// This type wraps `*mut xcb_connection_t` that is provided by libxcb. It provides a rust
 /// interface to this C library.
+#[allow(clippy::upper_case_acronyms)]
 #[derive(Debug)]
 pub struct XCBConnection {
-    conn: raw_ffi::XCBConnectionWrapper,
+    conn: raw_ffi::XcbConnectionWrapper,
     setup: Setup,
     ext_mgr: Mutex<ExtensionManager>,
     errors: pending_errors::PendingErrors,
@@ -68,7 +70,7 @@ impl XCBConnection {
             EXT_NOTSUPPORTED => ConnectionError::UnsupportedExtension,
             MEM_INSUFFICIENT => ConnectionError::InsufficientMemory,
             REQ_LEN_EXCEED => ConnectionError::MaximumRequestLengthExceeded,
-            FDPASSING_FAILED => ConnectionError::FDPassingFailed,
+            FDPASSING_FAILED => ConnectionError::FdPassingFailed,
             _ => ConnectionError::UnknownError,
             // Not possible here: PARSE_ERR, INVALID_SCREEN
         }
@@ -81,7 +83,7 @@ impl XCBConnection {
         match error {
             ERROR => IOError::new(ErrorKind::Other, ConnectionError::UnknownError).into(),
             MEM_INSUFFICIENT => ConnectError::InsufficientMemory,
-            PARSE_ERR => ConnectError::DisplayParsingError,
+            PARSE_ERR => DisplayParsingError::Unknown.into(),
             INVALID_SCREEN => ConnectError::InvalidScreen,
             _ => ConnectError::UnknownError,
             // Not possible here: EXT_NOTSUPPORTED, REQ_LEN_EXCEED, FDPASSING_FAILED
@@ -98,7 +100,7 @@ impl XCBConnection {
         unsafe {
             let mut screen: c_int = 0;
             let dpy_ptr = dpy_name.map_or(null(), |s| s.as_ptr());
-            let connection = raw_ffi::XCBConnectionWrapper::new(
+            let connection = raw_ffi::XcbConnectionWrapper::new(
                 raw_ffi::xcb_connect(dpy_ptr, &mut screen),
                 true,
             );
@@ -123,6 +125,8 @@ impl XCBConnection {
     /// Create a connection wrapper for a raw libxcb `xcb_connection_t`.
     ///
     /// `xcb_disconnect` is called on drop only if `should_drop` is `true`.
+    /// If this function returns an `Err()` and `should_drop` was true, then
+    /// `xcb_disconnect` was already called.
     ///
     /// # Safety
     ///
@@ -134,9 +138,10 @@ impl XCBConnection {
         should_drop: bool,
     ) -> Result<XCBConnection, ConnectError> {
         let ptr = ptr as *mut raw_ffi::xcb_connection_t;
+        let conn = raw_ffi::XcbConnectionWrapper::new(ptr, should_drop);
         let setup = raw_ffi::xcb_get_setup(ptr);
         Ok(XCBConnection {
-            conn: raw_ffi::XCBConnectionWrapper::new(ptr, should_drop),
+            conn,
             setup: Self::parse_setup(setup)?,
             ext_mgr: Default::default(),
             errors: Default::default(),
@@ -158,7 +163,7 @@ impl XCBConnection {
         let length = usize::from(length) * 4 + 8;
 
         let slice = from_raw_parts(wrapper.as_ptr(), length);
-        let result = Setup::try_parse(&*slice)?.0;
+        let result = Setup::try_parse(slice)?.0;
 
         Ok(result)
     }
@@ -198,7 +203,7 @@ impl XCBConnection {
             count: new_bufs.len(),
             ext: null_mut(), // Not needed since we always use raw
             opcode: 0,
-            isvoid: if has_reply { 0 } else { 1 },
+            isvoid: u8::from(!has_reply),
         };
         let mut flags = raw_ffi::send_request_flags::RAW;
         assert!(has_reply || !reply_has_fds);
@@ -292,7 +297,9 @@ impl XCBConnection {
 
     unsafe fn wrap_reply(&self, reply: *const u8, sequence: SequenceNumber) -> CSlice {
         // Update our "max sequence number received" field
-        atomic_u64_max(&self.maximum_sequence_received, sequence);
+        let _ = self
+            .maximum_sequence_received
+            .fetch_max(sequence, Ordering::Relaxed);
 
         let header = CSlice::new(reply, 32);
 
@@ -307,7 +314,9 @@ impl XCBConnection {
 
     unsafe fn wrap_error(&self, error: *const u8, sequence: SequenceNumber) -> CSlice {
         // Update our "max sequence number received" field
-        atomic_u64_max(&self.maximum_sequence_received, sequence);
+        let _ = self
+            .maximum_sequence_received
+            .fetch_max(sequence, Ordering::Relaxed);
 
         CSlice::new(error, 32)
     }
@@ -466,7 +475,10 @@ impl RequestConnection for XCBConnection {
 
         // The number of FDs is in the second byte (= buffer[1]) in all replies.
         let fd_slice = unsafe { std::slice::from_raw_parts(fd_ptr, usize::from(buffer[1])) };
-        let fd_vec = fd_slice.iter().map(|&fd| RawFdContainer::new(fd)).collect();
+        let fd_vec = fd_slice
+            .iter()
+            .map(|&fd| unsafe { OwnedFd::from_raw_fd(fd) })
+            .collect();
 
         Ok(ReplyOrError::Reply((buffer, fd_vec)))
     }
@@ -583,18 +595,17 @@ impl AsRawFd for XCBConnection {
     }
 }
 
-/// Atomically sets `value` to the maximum of `value` and `new`.
-fn atomic_u64_max(value: &AtomicU64, new: u64) {
-    // If only AtomicU64::fetch_max were stable...
-    let mut old = value.load(Ordering::Relaxed);
-    loop {
-        if old >= new {
-            return;
-        }
-        match value.compare_exchange_weak(old, new, Ordering::Relaxed, Ordering::Relaxed) {
-            Ok(_) => return,
-            Err(val) => old = val,
-        }
+#[cfg(unix)]
+impl AsFd for XCBConnection {
+    fn as_fd(&self) -> BorrowedFd<'_> {
+        unsafe { BorrowedFd::borrow_raw(self.as_raw_fd()) }
+    }
+}
+
+// SAFETY: We provide a valid xcb_connection_t that is valid for as long as required by the trait.
+unsafe impl as_raw_xcb_connection::AsRawXcbConnection for XCBConnection {
+    fn as_raw_xcb_connection(&self) -> *mut as_raw_xcb_connection::xcb_connection_t {
+        self.get_raw_xcb_connection().cast()
     }
 }
 
@@ -640,7 +651,7 @@ fn reconstruct_full_sequence_impl(recent: SequenceNumber, value: u32) -> Sequenc
 
 #[cfg(test)]
 mod test {
-    use super::{atomic_u64_max, XCBConnection};
+    use super::XCBConnection;
     use std::ffi::CString;
 
     #[test]
@@ -653,27 +664,6 @@ mod test {
         assert_eq!(screen, 0);
     }
 
-    #[test]
-    fn u64_max() {
-        use std::sync::atomic::{AtomicU64, Ordering};
-        let value = AtomicU64::new(0);
-
-        atomic_u64_max(&value, 3);
-        assert_eq!(value.load(Ordering::Relaxed), 3);
-
-        atomic_u64_max(&value, 7);
-        assert_eq!(value.load(Ordering::Relaxed), 7);
-
-        atomic_u64_max(&value, 5);
-        assert_eq!(value.load(Ordering::Relaxed), 7);
-
-        atomic_u64_max(&value, 7);
-        assert_eq!(value.load(Ordering::Relaxed), 7);
-
-        atomic_u64_max(&value, 9);
-        assert_eq!(value.load(Ordering::Relaxed), 9);
-    }
-
     #[test]
     fn reconstruct_full_sequence() {
         use super::reconstruct_full_sequence_impl;
